#!/bin/bash
# This script is used for creating edge-db PostgreSQL cluster

# should be mounted before running:
#
# 	mount proc -t proc /proc
# 	mount -L edgestor /var/opt/edge
# 	mount -L edgelog /var/log/edge
# @TODO: maybe some others mount needed

pushd `dirname $0` > /dev/null

# Available params, ./make_cluster.sh -?
_usage() {
    echo "Usage: $0 [-v version] [-c dbname] [-u username] [-p password] [-P port] [-d datadir] [-l logdir]"
    echo " NB! For default values see config/*.config!"
    echo "   -C, --config         Configuration name"
    echo "   -f, --force      Drop existing autocode cluster?"
    echo "   -v, --version    PostgreSQL version"
    echo "   -c, --cluster    PostgreSQL cluster name"
    echo "   -u, --user       PostgreSQL user name"
    echo "   -p, --password   PostgreSQL user password"
    echo "   -P, --port       PostgreSQL port"
    echo "   -d, --data-dir   PostgreSQL data directory"
    echo "   -l, --logs-dir   PosthreSQL logs directory"
    echo "   -?, --help       This help text"
    exit 0
}

# Current dir
_ME=$(readlink -f "$0")
PWD=$(dirname "${_ME}")
DROP_CLUSTER=false

# Config name
[ -f ${PWD}/.configure ] && CONFIG=$(cat ${PWD}/.configure | tr -d "\n\r")
CONFIG=${CONFIG:-default}

# CLI params
while test $# -gt 0; do
    case $1 in
        --config|-C)
            CONFIG="${2}"
            shift
        ;;
        --version|-v)
            PGVERSION="${2}"
            shift
        ;;
        --cluster|-c)
            PGCLUSTER="${2}"
            shift
        ;;
        --user|-u)
            PGUSER="${2}"
            shift
        ;;
        --password|-p)
            PGUSERPASS="${2}"
            shift
        ;;
        --port|-P)
            PGCLUSTERPORT="${2}"
            shift
        ;;
        --datadir|-d)
            PGDATADIR="${2}"
            shift
        ;;
        --logsdir|-l)
            PGLOGDIR="${2}"
            shift
        ;;
        --force|-f)
            DROP_CLUSTER=true
        ;;
        --help|-?)
            _usage
            break
        ;;

        --)
            break
        ;;
        --*|-?)
            echo "Unknown option ${1}"
            exit 11
        ;;
        -*)
            split=$1
            shift
            set -- $(echo "$split" | cut -c 2- | sed 's/./-& /g') "$@"
            continue
        ;;
        *)
            echo "Unknown argument ${1}"
            exit 11
        ;;
    esac

    shift
done

# Configs
source $PWD/config/$CONFIG.config

# Helpers
_error(){
    echo "ERROR: $*" >&2
    kill $$
    exit -1
}

# User must be root
if test "$EUID" -ne 0 ; then
    _error "This script MUST be run as root"
fi

# Main body
echo "Installing pgcluster..."

if [ ! -d /usr/share/fonts/truetype/dejavu ]; then
    TTDIR=$(find /usr/share/fonts/ -name "DejaVuSans.ttf" |sed 's/\/[^\/]*$//')
    [ -z "$TTDIR" ] && _error " finding TTF DejaVu fonts"
    ln -s $TTDIR /usr/share/fonts/truetype/dejavu
fi

PGUID=$(grep postgres /etc/passwd | cut -d: -f3)
PGGID=$(grep postgres /etc/passwd | cut -d: -f4)

# Checking current clusters and drop them if needed
CURRENT_CLUSTERS=$(pg_lsclusters | tail -n +2 | awk 'BEGIN{ str=""; }{ str=str$2" "str; }END{ print str; }')
if (echo "$CURRENT_CLUSTERS" | grep "$PGCLUSTER" > /dev/null); then
    if ((${DROP_CLUSTER})); then
	echo -n "WARNING: Autocode cluster to delete..."
	pg_dropcluster --stop $PGVERSION $PGCLUSTER

	if [ $? -ne 0 ]; then
	    _error "Cannot drop autocode cluster!"
	fi

	echo " done!"
    else
	echo "WARNING! CLUSTER $PGCLUSTER ALREADY EXISTS! SKIPPING CLUSTER PROCEDURE!"
	# Patching pgcluster port
	sed -i "s/\(port *= *\).*/\1$PGCLUSTERPORT/" /etc/postgresql/$PGVERSION/$PGCLUSTER/postgresql.conf
	pg_ctlcluster $PGVERSION $PGCLUSTER restart
	exit 0
    fi
fi

# Drop main cluster if exists
if (echo "$CURRENT_CLUSTERS" | grep main > /dev/null); then
    echo -n "WARNING: Main cluster to delete..."
    pg_dropcluster --stop $PGVERSION main

    if [ $? -ne 0 ]; then
	_error "Cannot drop autocode cluster!"
    fi

    echo " done!"
fi

if ! [ -d $PGDATADIR ]; then mkdir -p $PGDATADIR; fi
chown -R $PGUID:$PGGID $PGDATADIR
chmod 0755 $PGDATADIR
if ! [ -d $PGLOGDIR ]; then mkdir -p $PGLOGDIR; fi
chown -R $PGUID:$PGGID $PGLOGDIR
chmod 0755 $PGLOGDIR

pg_createcluster -d $PGDATADIR -l $PGLOGDIR/postgres-$PGVERSION-$PGCLUSTER.log --locale='C.UTF-8' -e 'UTF-8' -p $PGCLUSTERPORT --start-conf=auto \
    -o listen_addresses='*' \
    -o checkpoint_timeout=5min \
    -o enable_bitmapscan=on \
    -o enable_hashagg=on \
    -o enable_hashjoin=on \
    -o enable_indexscan=on \
    -o enable_material=on \
    -o enable_mergejoin=on \
    -o enable_nestloop=on \
    -o enable_seqscan=off \
    -o enable_sort=on \
    -o enable_tidscan=on \
    -o geqo=on \
    -o client_min_messages=error \
    -o log_min_messages=error \
    -o log_min_error_statement=error \
    -o log_line_prefix='%t [%p-%l] %q%u@%d ' \
    -o autovacuum=on \
    -o bytea_output='escape' \
    -o datestyle='iso, dmy' \
    -o array_nulls=on \
    -o log_timezone='localtime' \
    -o timezone='localtime' \
    $PGVERSION $PGCLUSTER

if [ $? -ne 0 ]; then
    _error "Cannot create autocode cluster!"
fi

printf "local\tall\t\tpostgres\t\t\ttrust\nlocal\tall\t\tall\t\t\t\ttrust\nhost\tall\t\tall\t\t127.0.0.1/32\ttrust\nhost\tall\t\tall\t\t::1/128\t\ttrust\n#host\tall\t\tall\t\t0.0.0.0/0\tmd5 # uncomment to open remote access\n" > \
    /etc/postgresql/$PGVERSION/$PGCLUSTER/pg_hba.conf

echo " done."

echo "Starting new autocode cluster..."

pg_ctlcluster $PGVERSION $PGCLUSTER start
pg_lsclusters | grep $PGVERSION | grep $PGCLUSTER | grep -q online || _error "failed to start postgresql!"

echo "Autocode Cluster ready!"

popd > /dev/null
exit 0
