--
--
-- api argument to be json with specific header object
-- { "@header":{"vendor":"VIT", "class":"vitml.edge.api", "version":"0.1", "object":"who", "action":"what"}  }
--
--


create or replace function api_create_header( iobject text, iaction text ) returns json as $$
	select row_to_json(t) from (
		select 
			_o('api_header_vendor') as vendor, 
			_o('api_header_class') as class, 
			_o('api_header_version') as version,
			to_json($1) as object,
			to_json($2) as action
	) t;
$$ language sql immutable;	

create or replace function api_valid_header( idata json, iobject varchar, iaction varchar ) returns boolean as $$
	select (
		(($1 -> '@header') notnull 
		and ($1 -> '@header' ->> 'vendor') = _o('api_header_vendor')
		and ($1 -> '@header' ->> 'class' ) = _o('api_header_class')) 
		and	($2 isnull or ($1 -> '@header' ->> 'object') = $2) 
		and ($3 isnull or ($1 -> '@header' ->> 'action') = $3)
	);
$$ language sql immutable;


--
-- { ... "guid": , "name": , "about": , "param": { ... }}
--

create or replace function api_post_add (idata json) returns json as $$
declare
	_guid uuid	:= (idata ->> 'guid')::uuid;
	_name text 	:= idata ->> 'name';
	_about text	:= idata ->> 'about';
	_param json := idata -> 'param';
begin
	if (not api_valid_header(idata,'post','add')) then
		return '{"errcode":-1, "errmsg":"Invalid header"}';
	end if;
	if _guid isnull then
		return '{"errcode":-1, "errmsg":"Invalid post GUID"}';
	end if;	
	if exists (select 1 from post where guid = _guid) then
		update post set
			name  = coalesce(_name, name) ,
			about = coalesce(_about, about),
			param = coalesce(_param, param)
		where guid = _guid;
	else
		execute format(
			'insert into post (guid,name,about,param) values ( $1, $2, $3, $4 )'
		) using _guid, _name, _about, _param;	
	end if;
	return json(format('{"errcode": "0", "errmsg": "Ok", "guid": %s}',to_json(_guid)));
exception when others then
	return json(format('{"errcode": "%s", "errmsg": %s}',SQLSTATE,to_json(SQLERRM)));
end;
$$ language plpgsql;


--
-- { ... "post": , "channel": , "name": , "description": , "param": { ... }}
--

create or replace function api_channel_add (idata json) returns json as $$
declare
	_post uuid			:= (idata ->> 'post')::uuid;
	_channel integer	:= (idata ->> 'channel')::integer;
	_name text 			:= idata ->> 'name';
	_description text	:= idata ->> 'description';
	_param json 		:= idata ->  'param';
	_id integer;
begin	
	if (not api_valid_header(idata,'channel','add')) then
		return '{"errcode":"-1", "errmsg": "Invalid header"}';
	end if;	
	if _post isnull or _channel isnull then
		return '{"errcode":"-1", "errmsg": "Invalid post and channel IDs"}';
	end if;	
	if exists (select 1 from channel where post = _post and channel = _channel) then
		update channel set
			name  = coalesce(_name, name),
			description = coalesce(_description, description),
			param = coalesce(_param, param)
		where post = _post and channel = _channel
		returning id into _id;	
	else
		execute format(
			'insert into channel (post,channel,name,description,param) values ( $1, $2, $3, $4, $5 ) returning id'
		) using _post, _channel, _name, _description, _param into _id;
	end if;	
	return json(format('{"errcode": "0", "errmsg": "Ok", "id": "%s"}',to_json(_id)));
exception when others then
	return json(format('{"errcode": "%s", "errmsg": %s}',SQLSTATE,to_json(SQLERRM)));
end;
$$ language plpgsql;



-- Function: api_plate_add(json)

-- DROP FUNCTION api_plate_add(json);

CREATE OR REPLACE FUNCTION api_plate_add(idata json)
  RETURNS json AS
$BODY$
declare
	_post uuid;
	_channel integer;
	_guid uuid;
    _lat float8;
    _lon float8;
	_id bigint;
	_sid bigint;
    _alarm integer;
begin	
	if (not api_valid_header(idata,'plate','add')) then
		return json(format('{"errcode": "%s", "errmsg": "%s"}',-1,'Invalid header'));
	end if;
	_guid := (idata ->> 'guid')::uuid;
	_post := (idata ->> 'post')::uuid;
	_channel := (idata ->> 'channel')::integer;
    _lat := (idata -> 'geodata' ->> 'latitude')::float8;
    _lon := (idata -> 'geodata' ->> 'longitude')::float8;
	if _guid isnull then
		return json(format('{"errcode": "%s", "errmsg": "%s"}',-2,'Invalid GUID'));
	end if;	
	if _post isnull then
		return json(format('{"errcode": "%s", "errmsg": "%s"}',-2,'Invalid POST'));
	end if;	
	if _channel isnull then
		return json(format('{"errcode": "%s", "errmsg": "%s"}',-2,'Invalid CHANNEL'));
	end if;	
	if _o('platear_except_unknown_post')::boolean and not exists (select 1 from public.post where guid = _post) then
        return json(format('{"errcode": "%s", "errmsg": "%s"}',11,'Unknown POST'));
    end if;
	if _o('platear_except_unknown_channel')::boolean and not exists (select 1 from public.channel where post = _post and channel = _channel) then
        return json(format('{"errcode": "%s", "errmsg": "%s"}',12,'Unknown CHANNEL'));
    end if;
    if exists (select 1 from platear where guid = _guid) then
		return json(format('{"errcode": "%s", "errmsg": "%s"}',1,'GUID already exists'));
	end if;
    if _lat notnull and _lon notnull then
        update channel set name = _lat::text, description = _lon::text where post = _post and channel = _channel;
    end if;    
    select into _alarm
        coalesce(sum(c.value),0) 
    from  
        json_array_elements_text(idata->'alarm') alarm 
        join const_integer c on c.name = alarm and c.section = 'mask.issue.type';

    execute format(
        'insert into platear 
        (
            guid,
            ctime,
            stime,
            etime,
            post_id,
            channel,
                rect_left,
                rect_top,
                rect_right,
                rect_bottom,
            "number",
            country,
            county,
            numtype,
            eventspec,
            direction,
            geodata,
            velocity,
            validity,
            int0,
            vehicle_length,
            lane,
            alarm_type
         )
        values (
            $1    
            ,%s
			,%s
			,%s
            ,$2
            ,$3
                ,$4,$5,$6,$7
            ,$8    
            ,%s
            ,%s
            ,%s
            ,$9
            ,$10
            ,$11
            ,%s
            ,%s
            ,$12
            ,%s
            ,%s
            ,$13
        ) 
        returning id, scrollid'
        ,coalesce(''''||lib.us2ts((idata ->> 'ctime')::bigint)::text||'''','default')       
        ,coalesce(''''||lib.us2ts((idata ->> 'stime')::bigint)::text||'''',coalesce(''''||lib.us2ts((idata ->> 'ctime')::bigint)::text||'''','default'))       
        ,coalesce(''''||lib.us2ts((idata ->> 'etime')::bigint)::text||'''',coalesce(''''||lib.us2ts((idata ->> 'ctime')::bigint)::text||'''','default'))       
		,coalesce((idata ->> 'country')::integer::text,'default')
		,coalesce((idata ->> 'county')::integer::text,'default')
		,coalesce((idata ->> 'numtype')::integer::text,'default')
		,coalesce(round((idata ->> 'velocity')::float)::integer::text, 'default')
		,coalesce(round((idata ->> 'validity')::float)::integer::text, 'default')
		,coalesce((idata ->> 'vehicle_length')::integer::text, 'default')
		,coalesce((idata ->> 'lane')::integer::text, 'default')
    )
    using 
        _guid
        ,_post
        ,_channel
            ,(idata -> 'rect' ->> 'left')::float8
            ,(idata -> 'rect' ->> 'top')::float8
            ,(idata -> 'rect' ->> 'right')::float8
            ,(idata -> 'rect' ->> 'bottom')::float8
        ,(idata ->> 'number')
        ,0
        ,(idata ->> 'direction')::integer
        ,format('(%s,%s,%s,%s,%s,%s,%s)',   
            (idata -> 'geodata' ->> 'latitude')::float8,
            (idata -> 'geodata' ->> 'longitude')::float8,
            (idata -> 'geodata' ->> 'altitude')::float8,
            (idata -> 'geodata' ->> 'roll')::float8,
            (idata -> 'geodata' ->> 'yaw')::float8,
            (idata -> 'geodata' ->> 'pitch')::float8,
            (idata -> 'geodata' ->> 'speed')::float8
        )::geodata
		,round((idata ->> 'overspeed')::float)::integer
        ,_alarm
    into _id, _sid;
    
    if _id isnull or _sid isnull then 
		return json(format('{"errcode": "2", "errmsg": %s}',
			to_json('invalid insert -> id: %s sid: %s',
				coalesce(_id::text,'<null>'),
					coalesce(_sid::text,'<null>')))); 
	end if;
	return json(format('{"errcode": "0", "errmsg": "Ok", "id": %s, "scrollid": %s}',_id,_sid));
exception when others then
	return json(format('{"errcode": "%s", "errmsg": %s}',SQLSTATE,to_json(SQLERRM)));
end;
$BODY$ language plpgsql;

--
-- { "object":"post", "action":"discovery" }
--
-- { "object":"channel", "action":"discovery",... "post": }
--
-- { "object":"string", "action":"discovery",... "section":..."lang":... }
--
-- { "object":"opt","action":"discovery",...,"section":...,"keys":...,"lang":... }
--
-- { "object":"const","action":"discovery","kind":["int","text"],"section",...}
--
-- { "object":"mediastore", "action":"discovery", "post":, "channel":, ...}
--
create or replace function api_discovery (idata json) returns json as $$
declare
	_post uuid;
    _channel integer;
    _section text;
	_keys text;
	_kind text;
    _lang varchar(2);
	_res json;
begin
	if (api_valid_header(idata,'post','discovery')) then
		select into _res row_to_json(rr) from(select '0' as errcode, '' as errmsg, json_agg(row_to_json(r)) as posts from (select *,(select json_agg(row_to_json(c)) from (select channel,name,description,param from channel where post = p.guid) c) as channels from post p) r) rr;
	elsif (api_valid_header(idata,'channel','discovery')) then
		_post := (idata ->> 'post')::uuid;
		select into _res row_to_json(rr) from (select '0' as errcode, '' as errmsg, json_agg(row_to_json(r)) as channels from (select post,channel,name,description,param from channel where post = _post) r) rr;
	elsif (api_valid_header(idata,'string','discovery')) then
        _section := (idata ->> 'section');
        _lang := coalesce((idata ->> 'lang'),(_o('language_priority')::text[])[1]);
        select into _res 
            row_to_json(rr) 
        from ( select '0' as errcode, '' as errmsg, json_agg(r) as strings
            from (
                select 
                    key as value, 
                    value as caption 
                from 
                    public.stringtable 
                where section = _section and lang = _lang
            ) r 
        ) rr;
    elsif (api_valid_header(idata,'opt','discovery')) then
        _section := (idata ->> 'section');
        _keys := (idata ->> 'keys');
        _lang := coalesce((idata ->> 'lang'),(_o('language_priority')::text[])[1]);
        select into _res 
            row_to_json(rr) 
        from ( select '0' as errcode, '' as errmsg, json_agg(r) as opt
            from (
				select 
					o.section,o.key,o.value,o.default, o.type, o.group, o.weight,
					s.value as description 
				from 
					core.opt o 
					left join public.stringtable s 
						on s.section= 'opt' and 
							s.lang = _lang and 
							s.key = _section||'.'||o.key 
				where 
					case 
						when _section isnull or _section = '' 
						then true
						else o.section = _section 
					end
					and 
					case 
						when (_keys isnull or _keys = '') 
						then true 
						else o.key ~ _keys
					end	
            ) r 
        ) rr;
    elsif (api_valid_header(idata,'const','discovery')) then
        _section := (idata ->> 'section');
        _kind := (idata ->> 'kind');
		if _kind isnull or _kind = '' then
			select into _res 
				row_to_json(rr) 
			from ( select '0' as errcode, '' as errmsg, 
				( select json_object_agg(name, value)
					from public.const_integer
					where case when _section isnull or _section = '' then true else section = _section end
				) as const_int,	
				( select json_object_agg(name, value)
					from public.const_text
					where case when _section isnull or _section = '' then true else section = _section end 
				) as const_text	
			) rr;
		elsif _kind = any(array['int','integer']) then
			select into _res 
				row_to_json(rr) 
			from ( select '0' as errcode, '' as errmsg, 
				( select json_object_agg(name, value)
					from public.const_integer
					where section = _section 
				) as const_int	
			) rr;
		elsif _kind = any(array['text','string']) then 
			select into _res 
				row_to_json(rr) 
			from ( select '0' as errcode, '' as errmsg, 
				( select json_object_agg(name, value)
					from public.const_text
					where section = _section 
				) as const_text	
			) rr;
		else
			select into _res 
				row_to_json(rr) 
			from ( select '-2' as errcode, 'Invalid kind' as errmsg ) rr;
		end if;
	elsif (api_valid_header(idata,'mediastore','discovery')) then
        _post := (idata ->> 'post')::uuid;
        _channel := (idata ->> 'channel')::integer; 
        select into _res 
            row_to_json(rr) 
        from ( select '0' as errcode, '' as errmsg, json_agg(r) as mediastore
            from (
                select 
                    m.post, 
                    m.channel,
                    m.image, 
                    extract(epoch from m.minage)::text||'s' as minage, 
                    extract(epoch from m.maxage)::text||'s' as maxage, 
                    s.root,
                    s.class,
                    json(s.opt) as opt 
                from 
                    plate_mediastore m 
                        join binfs.store s on s.name = m.store
                where
                    ((_post is null) or (post = _post)) and
                    ((_channel is null) or (channel = _channel))
            ) r
        ) rr;
    else
		return '{"errcode":"-1", "errmsg": "Invalid header"}';
	end if;
	return _res;
exception when others then
	return json(format('{"errcode": "%s", "errmsg": %s}',SQLSTATE,to_json(SQLERRM)));
end;
$$ language plpgsql;


-- 
-- { "object":"plate", "action":"attrget", ... }
-- { ..., "post": (guid), "channel": [int], "plate": (string), "fromts": (long), "tots": (long), "distance": (int), "limit": (int)  "alarms": [(string)]}
--
create or replace function api_plateattrget ( idata json ) returns setof platear as $body$
declare
	i_post uuid;
	i_channel integer[];
	i_from_ts timestamptz;
	i_to_ts timestamptz;
	i_plate varchar;
	i_limit integer;
	i_distance integer;
	i_fromspeed integer;
	i_tospeed integer;
	i_alarms text;
	i_order text;
	regex varchar;
	sql text;
begin
	if not api_valid_header(idata,'plate','attrget') then
		perform core.log('api_plateattrget','error','invalid header');
		return;
	end if;
	
	i_post := idata ->> 'post';
	
	if json_typeof(idata -> 'channel') isnull then
		i_channel := '{}'::int[];
	elsif json_typeof(idata -> 'channel') = 'array' then 	
		i_channel := lib.json_array_int(idata -> 'channel');
	elsif json_typeof(idata -> 'channel') = 'number' then
		i_channel := array[(idata ->> 'channel')::int];
	else
		i_channel := '{}'::int[];
	end if;	

	i_from_ts := lib.us2ts((idata ->> 'fromts')::bigint);
	i_to_ts   := lib.us2ts((idata ->> 'tots')::bigint);
	if i_from_ts isnull then i_from_ts := '-infinity'::timestamptz; end if;
	if i_to_ts isnull   then i_to_ts   := ' infinity'::timestamptz; end if;
	
	i_distance := idata ->> 'distance';
	if i_distance isnull then i_distance := 0; end if;
	
	i_fromspeed := idata ->> 'fromspeed';
	i_tospeed := idata ->> 'tospeed';

    select 
        into i_alarms
        ' and (false '||string_agg(
            case 
                when mask = 0 then 'or alarm_type = 0' 
                when mask = -1 then 'or alarm_type > 0' 
                else format('or (alarm_type & %s) = %s',mask,mask) end,
            ' ')||') '
    from
    ( 
        select 
            name, 
            value as mask 
        from const_integer c 
        where c.section = 'mask.issue.type' 
        union all
        values ('_none' , 0),('_any' , -1)
    ) q
    join (
        select json_array_elements_text(idata -> 'alarms') as name
    ) p
    using (name);

	i_limit := idata ->> 'limit';
	if i_limit isnull then 
		i_limit := _o('platelist_default_limit')::integer; 
	end if;
	if i_limit > _o('platelist_maximum_limit')::integer then 
		perform core.log('api_plateattrget','error','maximum limit exceed');
		return;
	end if;

	i_order := idata ->> 'order';
	if (i_order isnull) then
		i_order := 'asc';
	ELSIF (lower(i_order) <>  'desc' AND lower(i_order) <>  'asc') then
		perform core.log('api_plateattrget','error','invalid order option');
	return;
	end if;
	
	i_plate := idata ->> 'plate';
	regex   := idata ->> 'plate';

	if( i_plate ~ E'[^\\.]\\*' or i_plate ~ E'^\\*' ) then
		regex := replace(i_plate,'*','.*');
	end if;

	if ( regex !~ E'^\\^' ) then
		regex := E'^'||regex;
	end if;

	if ( regex !~ E'\\$$' ) then
		regex := regex||E'$';
	end if;

	if i_plate ~ E'[\\[\\]\\^\\$\\\\\\.\\\\+\\*\\?]' then
		sql  := 'select * from platear
			where
			post_id = $1 and
			case when $2 = array[]::int[] then true else channel = any($2) end and
			number ~* $3 and
			$4 <= ctime and ctime < $5';
		if i_fromspeed notnull then sql := sql || ' and velocity >= $7 '; end if;  
		if i_tospeed   notnull then sql := sql || ' and velocity <= $8 '; end if;  
		if i_alarms    notnull then sql := sql || i_alarms; end if;  
		sql := sql || ' order by ctime '||i_order||' limit $6';
		return query
			execute sql 
			using i_post,i_channel,regex,i_from_ts,i_to_ts,i_limit,i_fromspeed,i_tospeed;
	elsif i_distance > 0 then
		sql := 'select * from platear
			where
			post_id = $1 and
			case when $2 = array[]::int[] then true else channel = any($2) end and
			lib.ilevenstein(number,$3) <= $4 and
			$5 <= ctime and ctime < $6 ';
		if i_fromspeed notnull then sql := sql || ' and velocity >= $8 '; end if;  
		if i_tospeed   notnull then sql := sql || ' and velocity <= $9 '; end if;  
		if i_alarms    notnull then sql := sql || i_alarms;  end if;
		sql := sql || ' order by ctime '||i_order||' limit $7';	
		return query
			execute sql 
			using i_post,i_channel,i_plate,i_distance,i_from_ts,i_to_ts,i_limit,i_fromspeed,i_tospeed;
	else
		sql := 'select * from platear
			where
			post_id = $1 and
			case when $2 = array[]::int[] then true else channel = any($2) end and
			number = $3 and
			$4 <= ctime and ctime < $5 '; 
		if i_fromspeed notnull then sql := sql || ' and velocity >= $7'; end if;
		if i_tospeed   notnull then sql := sql || ' and velocity <= $8'; end if;
		if i_alarms    notnull then sql := sql || i_alarms;  end if;
		sql := sql || ' order by ctime '||i_order||' limit $6';
		return query
			execute sql
			using i_post,i_channel,i_plate,i_from_ts,i_to_ts,i_limit,i_fromspeed,i_tospeed;
	end if;
end;
$body$ language plpgsql;

CREATE OR REPLACE FUNCTION api_channel_delete(idata json)
      RETURNS json AS
$BODY$
declare
    r plate_mediastore%rowtype;
    _post uuid;
    _channel integer;
    _root character varying(255);
begin
    if (not api_valid_header(idata,'channel','delete')) then
        return '{"errcode":"-1", "errmsg":"Invalid header"}';
    end if;
    _post := (idata ->> 'post')::uuid;
    _channel := (idata ->> 'channel')::integer;
    _root := (idata ->> 'root')::character varying(255);
    if exists (select 1 from channel where post = _post and channel = _channel) then
        --delete plate_mediastore
        FOR r IN SELECT * FROM plate_mediastore WHERE post = _post and channel = _channel
        LOOP
            perform api_plate_mediastore_delete(json('{"@header": '||api_create_header('plate_mediastore','delete')||',"post":"'||_post||'","channel":'||_channel||',"image":"'||r.image||'"}'));
        END LOOP;

        DELETE FROM platear
            WHERE post_id = _post and channel = _channel;
        DELETE FROM channel
            WHERE post = _post and channel = _channel;
        perform binfs.rmtree(lib.pathjoin(_root,_post::text,_channel::text));
        return json(format('{"errcode": "0", "errmsg": "Chanel Deleted!", "post": %s,"channel":%s}',to_json(_post),to_json(_channel)));
    else
        return json(format('{"errcode": "0", "errmsg": "Chanel not found!", "post": %s,"channel":%s}',to_json(_post),to_json(_channel)));
    end if;
exception when others then
    return json(format('{"errcode": "%s", "errmsg": %s}',SQLSTATE,to_json(SQLERRM)));
end;
$BODY$ LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION api_plate_mediastore_delete(idata json)
RETURNS json AS
$BODY$
declare
    _post uuid;
    _channel integer;
    _image varchar;
    _store varchar;
	_res integer;
begin
    if (not api_valid_header(idata,'plate_mediastore','delete')) then
        return '{"errcode":"-1", "errmsg":"Invalid header"}';
    end if;
    _post := (idata ->> 'post')::uuid;
    _channel := (idata ->> 'channel')::integer;
    _image := (idata ->> 'image')::character varying(255);
    
	for _store in select store from plate_mediastore where post = _post and channel = _channel and image = _image loop
		if _store notnull then
			select into _res binfs._remove(_store);
			select into _res binfs._remove('META..'||_store);
			delete from plate_mediastore where store = _store;
		end if;
	end loop;	
	return json(format('{"errcode": "0", "errmsg": "Plate_mediastore Deleted!", "post": %s,"channel":%s,"image":%s}',to_json(_post),to_json(_channel),to_json(_image)));

exception when others then
    return json(format('{"errcode": "%s", "errmsg": %s}',SQLSTATE,to_json(SQLERRM)));
end;
$BODY$ LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION api_post_delete(idata json) RETURNS json AS
$BODY$
declare
    _root core.opt.value%type;
    r channel%rowtype;
    _post uuid;
begin
    if (not api_valid_header(idata,'post','delete')) then
        return '{"errcode":"-1", "errmsg":"Invalid header"}';
    end if;
    _post := (idata ->> 'guid')::uuid;
    _root := _o('plateimage_store_root');
    if exists (SELECT 1 FROM post WHERE guid = _post) then
        FOR r IN SELECT * FROM channel WHERE post = _post
        LOOP
            perform api_channel_delete(json('{"@header": '||api_create_header('channel','delete')||',"post":"'||_post||'", "channel":'||r.channel||',"root":"'||_root||'"}'));
        END LOOP;
        
        DELETE FROM post WHERE guid = _post;
        perform binfs.rmtree(lib.pathjoin(_root,_post::text));
        
        return json(format('{"errcode": "0", "errmsg": "Post Deleted!", "guid": %s}',to_json(_post)));
    else
        return json(format('{"errcode": "0", "errmsg": "Post not found!", "guid": %s}',to_json(_post)));
        
    end if;
exception when others then
    return json(format('{"errcode": "%s", "errmsg": %s}',SQLSTATE,to_json(SQLERRM)));
end;
$BODY$ LANGUAGE plpgsql;


--
-- { "object":"report", "action":"get" }
-- { ..., "name":"", "param":{...} }
--
create or replace function api_report_get (idata jsonb) returns jsonb as $$
declare
    _name varchar  := idata ->> 'name';
    _param jsonb := coalesce(idata -> 'param', '{}'::jsonb);      
	_handler varchar;
    _res jsonb;
begin
	if (api_valid_header(idata::json,'report','get')) then
	    if _name isnull then
	    	return '{"errcode":-1, "errmsg":"Invalid report name"}';
	    end if;	
	    select into _handler handler from public.handlers where section = 'report' and handle = _name and enabled;
        if not found then
	    	return '{"errcode":1, "errmsg":"Valid report handler not found"}';
        end if;
		execute format('select %s ($1)', _handler) using _param into _res;
    else
		return '{"errcode":"-1", "errmsg": "Invalid header"}';
	end if;
	return _res;
exception when others then
	return json(format('{"errcode": "%s", "errmsg": %s}',SQLSTATE,to_json(SQLERRM)));
end;
$$ language plpgsql;

