
--
-- report handler on (plate.scurry)
-- PARAM:
--      post     guid               - null for any
--      channel  int                - null for any
--      fromtime timestamp in us    - time domain start (def: -infinit)
--      totime   timestamp in us    - time domain end (def: infinit)
--      mingap   int                - gap in seconds between separate detections (def: 10)
--      mindet   int                - minimal number of detections to be reported (def: 2)
-- RETURNS: jsonb object as
--      errcode int (def:0)
--      errmsg  varchar (def:Ok)
--      result:
--          [{"total-count","plate-text","last-guid","last-datetime","last-ts"}] - frequency
--          analysis of plate detections, sorted by scurry frequency
--

create or replace function public.report_plate_scurry( iparam jsonb ) returns jsonb as $$
declare
    ipost     uuid;
    ichannel  int;
    ifromtime timestamptz;
    itotime   timestamptz;
    imingap   int;
    imindet   int;
    result   jsonb;
begin
    ipost := (iparam ->> 'post')::uuid;
    ichannel := (iparam ->> 'channel')::int;
    ifromtime := coalesce(lib.us2ts((iparam ->> 'fromtime')::bigint), '-infinity'::timestamptz);
    itotime   := coalesce(lib.us2ts((iparam ->> 'totime')::bigint),    'infinity'::timestamptz);
    imingap := coalesce((iparam ->> 'mingap')::integer, 10);
    imindet := coalesce((iparam ->> 'mindet')::integer, 2);
    select into result json_agg(r)::jsonb from (
        select plate AS "plate-text",
            count AS "total-count",
            lastguid AS "last-guid",
            lasttime AS "last-datetime",
            lib.ts2us(lasttime) AS "last-ts"
        from (
            select distinct
                q2.number as plate,
                count(number) over (partition by number) as count,
                first_value(ctime) over (partition by number) as lasttime,
                first_value(guid) over (partition by number) as lastguid
            from (
                select
                    q1.*,
                    not (number = n2 and t2 notnull and abs(extract(epoch from t2-ctime)) < imingap) as outgap
                from (
                    select
                        guid,
                        ctime,
                        number,
                        lag(ctime,1) over(order by number,ctime desc) as t2,
                        lag(number,1) over(order by number,ctime desc) as n2
                    from platear
                    where true
                        and (ipost isnull or (ipost notnull and post_id = ipost))
                        and (ichannel isnull or (ichannel notnull and channel = ichannel))
                        and ctime > ifromtime and ctime < itotime
                    order by number,ctime desc
                ) q1
             ) q2
            where outgap
        ) q3
        where count >= imindet
        order by count desc
    ) r;

    return json_build_object('errcode',0,'errmsg','Ok','result',coalesce(result,'[]'::jsonb));

exception when others then
    
    return json(format('{"errcode": "%s", "errmsg": %s, "result":[]}',SQLSTATE,to_json(SQLERRM)));

end;
$$ language plpgsql;


--
-- report handler on (plate.traffic.combined)
-- PARAM:
--    ordinate          varchar as (none, speed, alarm, category, weight)
--    datetime-min      timestamp (null)
--    datetime-max      timestamp (null)
--    datetime-step     varchar as {hour, day, week, month, year}
--    post              uuid (null)
--    channel           int (null)
--    lane              int[] (null)
--    lane-stat         varchar as (undefined, defined)
--    plate-stat        varchar as (undefined, defined)
--    speed-stat        varchar as (undefined, defined)
--    speed-min         int (null)    
--    speed-max         int (null)
--    overspeed-min     int (null)
--    speed-step        int (null)
--    weight-min        int (null)
--    weight-max        int (null)
--    weight-step       int (null)
--    category          varchar[] as {}
--    alarm             varchar[] as { none, speedviolation, watchlist }
--RETURNS:
--    {"limits":, "count": , "items": [ {"datetime":, "count":, "items": [ {"$ordinate":, "count":}, ] }, ]}
--
--

create or replace function public.report_traffic_combined( iparam jsonb ) returns jsonb as $$
declare
    iordinate varchar;
    ifromtime timestamptz;
    itotime timestamptz;
    itimedelta varchar(5);
    ipost uuid;
    ichannel int;
    ilane text;
    ifromspeed int;
    itospeed int;
    ispeeddelta int;
    ioverspeed int;
    ifromweight int;
    itoweight int;
    iweightdelta int;
    ialarms text;   
    icategory text;
    iwithlane boolean;
    iwithplate boolean;
    iwithspeed boolean;
    _str text;
    _items jsonb;
    _total int;
begin
    iordinate := lower((iparam ->> 'ordinate'));
    if iordinate isnull or not iordinate = any(array['none', 'speed','alarm','weight','category']) then
        iordinate := 'none';
    end if;    
    ifromtime := (iparam ->> 'datetime-min')::timestamptz;
    itotime   := (iparam ->> 'datetime-max')::timestamptz;
    itimedelta := lower((iparam ->> 'datetime-step'));
    if itimedelta isnull or not itimedelta = any(array['none', 'hour', 'day', 'week', 'month', 'year']) then
        if itotime notnull and ifromtime notnull and itotime - ifromtime < interval '1 day' then
            itimedelta = 'hour';
        else
            itimedelta = 'day';
        end if;    
    end if;
    ipost := (iparam ->> 'post')::uuid;
    ichannel := (iparam ->> 'channel')::int;
    ilane := (iparam ->> 'lane');
    ifromspeed := (iparam ->> 'speed-min')::int;
    itospeed := (iparam ->> 'speed-max')::int;
    ioverspeed := (iparam ->> 'overspeed-min')::int;
    ispeeddelta := coalesce((iparam ->> 'speed-step')::int,20);
    ifromweight := (iparam ->> 'weight-min')::int;
    itoweight := (iparam ->> 'weight-max')::int;
    iweightdelta := coalesce((iparam ->> 'weight-step')::int,20);
    _str := lower((iparam ->> 'plate-stat'));
    if _str notnull and _str = any(array['defined','undefined']) then
        iwithplate := (_str = 'defined');
    end if;    
    _str := lower((iparam ->> 'speed-stat'));
    if _str notnull and _str = any(array['defined','undefined']) then
        iwithspeed := (_str = 'defined');
    end if;    
    _str := lower((iparam ->> 'lane-stat'));
    if _str notnull and _str = any(array['defined','undefined']) then
        iwithlane := (_str = 'defined');
    end if;    
    select 
        into ialarms
        'and (false '||string_agg(
            case 
                when mask = 0 then 'or alarm_type = 0' 
                when mask = -1 then 'or alarm_type > 0' 
                else format('or (alarm_type & %s) = %s',mask,mask) 
            end,
            ' ')||')'
    from
    ( 
        select 
            name, 
            value as mask 
        from const_integer c 
        where c.section = 'mask.issue.type' 
        union all 
        values ('_none', 0), ('_any', -1)
    ) q
    join (
        select jsonb_array_elements_text(iparam -> 'alarm') as name
    ) p
    using (name);
    icategory := null;
    if iordinate = 'none' then
        execute format ( '
            select
                json_agg(r)::jsonb, sum(r.count) 
            from (    
                select 
                    case when ''%s'' = ''none'' then null else date_trunc(''%s'', ctime) end as datetime, count(*) %s as count 
                from 
                    platear 
                where true
                    %s
                    %s
                    %s
                    %s
                    %s
                    %s
                    %s
                    %s
                    %s
                    %s
                    %s
                 group by datetime
                 order by datetime
             ) r    
            ', 
            itimedelta,itimedelta,
            case when ioverspeed notnull then format('filter where(int0 > %s)', ioverspeed) else null end,
            case when ifromtime notnull then format('and ctime >= ''%s''', ifromtime) else null end,
            case when itotime notnull then format('and ctime <= ''%s''', itotime) else null end,
            case when ipost notnull then format('and post_id = ''%s''', ipost) else null end,
            case when ichannel notnull then format('and channel = %s', ichannel) else null end,
            case when ilane notnull then format('and lane = any(array %s)', ilane) else null end,
            case when iwithlane notnull then format('and lane is %s null',case when iwithlane then 'not' else null end) else null end,
            case when iwithplate notnull then format('and number is %s null',case when iwithplate then 'not' else null end) else null end,
            case when iwithspeed notnull then format('and velocity is %s null',case when iwithspeed then 'not' else null end) else null end,
            case when ifromspeed notnull then format('and velocity >= %s', ifromspeed) else null end,
            case when itospeed notnull then format('and velocity <= %s', itospeed) else null end,
            ialarms)
        into _items, _total;    
    elsif iordinate = 'speed' then
        execute format ( '
            select
                json_agg(r)::jsonb, sum(r.count) 
            from (   
                select
                    time as datetime, 
                    sum(count) as count, 
                    json_agg(json_build_object(''speed'',rr.speed,''count'',rr.count)) as items
                from (    
                    select 
                        case when ''%s'' = ''none'' then null else date_trunc(''%s'', ctime) end as time, (velocity/%s)*%s as speed, count(*) %s as count 
                    from 
                        platear 
                    where true
                        %s
                        %s
                        %s
                        %s
                        %s
                        %s
                        %s
                        %s
                        %s
                        %s
                        %s
                     group by time, speed
                     order by time, speed
                 ) rr
                 group by time
                 order by time
             ) r    
            ', 
            itimedelta, itimedelta,
            ispeeddelta, ispeeddelta,
            case when ioverspeed notnull then format('filter where(int0 > %s)', ioverspeed) else null end,
            case when ifromtime notnull then format('and ctime >= ''%s''', ifromtime) else null end,
            case when itotime notnull then format('and ctime <= ''%s''', itotime) else null end,
            case when ipost notnull then format('and post_id = ''%s''', ipost) else null end,
            case when ichannel notnull then format('and channel = %s', ichannel) else null end,
            case when ilane notnull then format('and lane = any(array %s)', ilane) else null end,
            case when iwithlane notnull then format('and lane is %s null',case when iwithlane then 'not' else null end) else null end,
            case when iwithplate notnull then format('and number is %s null',case when iwithplate then 'not' else null end) else null end,
            case when iwithspeed notnull then format('and velocity is %s null',case when iwithspeed then 'not' else null end) else null end,
            case when ifromspeed notnull then format('and velocity >= %s', ifromspeed) else null end,
            case when itospeed notnull then format('and velocity <= %s', itospeed) else null end,
            ialarms)
        into _items, _total;   
    elsif iordinate = 'alarm' then
        execute format ( '
            select
                json_agg(r)::jsonb, sum(r.count) 
            from (   
                select
                    time as datetime, 
                    sum(count) as count, 
                    json_agg(json_build_object(''alarm'',rr.alarm,''count'',rr.count)) as items
                from (    
                    select 
                        case when ''%s'' = ''none'' then null else date_trunc(''%s'', ctime) end as time , coalesce(m.name,''_none'') as alarm, count(*) %s as count
                    from 
                        platear p left join (
                            select 
                                name, 
                                value as mask 
                            from const_integer 
                            where section = ''mask.issue.type''
                        ) m on (p.alarm_type & m.mask) = m.mask
                    where true
                        %s
                        %s
                        %s
                        %s
                        %s
                        %s
                        %s
                        %s
                        %s
                        %s
                     group by time, alarm
                     order by time, alarm
                 ) rr
                 group by time
                 order by time
             ) r    
            ', 
            itimedelta, itimedelta,
            case when ioverspeed notnull then format('filter where(int0 > %s)', ioverspeed) else null end,
            case when ifromtime notnull then format('and ctime >= ''%s''', ifromtime) else null end,
            case when itotime notnull then format('and ctime <= ''%s''', itotime) else null end,
            case when ipost notnull then format('and post_id = ''%s''', ipost) else null end,
            case when ichannel notnull then format('and channel = %s', ichannel) else null end,
            case when ilane notnull then format('and lane = any(array %s)', ilane) else null end,
            case when iwithplate notnull then format('and number is %s null',case when iwithplate then 'not' else null end) else null end,
            case when iwithspeed notnull then format('and velocity is %s null',case when iwithspeed then 'not' else null end) else null end,
            case when ifromspeed notnull then format('and velocity >= %s', ifromspeed) else null end,
            case when itospeed notnull then format('and velocity <= %s', itospeed) else null end,
            ialarms)
        into _items, _total;   
    elsif iordinate = 'category' then

    elsif iordinate = 'weight' then

    end if;

    if _items notnull and _total notnull then
        return json_build_object('errcode',0,'errmsg','Ok','result', json_build_object('limits',iparam,'count',_total,'items',_items)::jsonb);
    else    
        return json_build_object('errcode',0,'errmsg','Ok','result', json_build_object('limits',iparam,'count',0,'items','[]'::jsonb)::jsonb);
    end if;    

exception when others then
    return json(format('{"errcode": "%s", "errmsg": %s}', SQLSTATE, to_json(SQLERRM)));
end;
$$ language plpgsql;
